"""
Test all OSSFile related methods
"""
# pylint:disable=invalid-name
# pylint:disable=missing-function-docstring
# pylint:disable=protected-access
import io
import os

import pytest

from .conftest import LICENSE_PATH, NUMBERS


def test_simple(ossfs, test_path):
    file = test_path + "/test_simple/file"
    data = os.urandom(10 * 2**20)

    with ossfs.open(file, "wb") as f:
        f.write(data)

    with ossfs.open(file, "rb") as f:
        out = f.read(len(data))
        assert len(data) == len(out)
        assert out == data


def test_seek(ossfs, test_path):
    file = test_path + "/test_seek/file"
    with ossfs.open(file, "wb") as f:
        f.write(b"123")

    with ossfs.open(file) as f:
        f.seek(1000)
        with pytest.raises(ValueError):
            f.seek(-1)
        with pytest.raises(ValueError):
            f.seek(-5, 2)
        with pytest.raises(ValueError):
            f.seek(0, 10)
        f.seek(0)
        assert f.read(1) == b"1"
        f.seek(0)
        assert f.read(1) == b"1"
        f.seek(3)
        assert f.read(1) == b""
        f.seek(-1, 2)
        assert f.read(1) == b"3"
        f.seek(-1, 1)
        f.seek(-1, 1)
        assert f.read(1) == b"2"
        for i in range(4):
            assert f.seek(i) == i


def test_read_small(ossfs, number_file):
    with ossfs.open(number_file, "rb", block_size=3) as f:
        out = []
        while True:
            data = f.read(2)
            if data == b"":
                break
            out.append(data)
        assert ossfs.cat(number_file) == b"".join(out)


def test_read_ossfs_block(ossfs, license_file, number_file):
    with open(LICENSE_PATH, "rb") as f_r:
        data = f_r.read()
    lines = io.BytesIO(data).readlines()
    assert ossfs.read_block(license_file, 0, 10, b"\n") == lines[0]
    assert ossfs.read_block(license_file, 40, 10, b"\n") == lines[1]
    assert ossfs.read_block(license_file, 0, 80, b"\n") == lines[0] + lines[1]
    assert ossfs.read_block(license_file, 0, 120, b"\n") == b"".join(
        [lines[0], lines[1], lines[2]]
    )

    lines = io.BytesIO(NUMBERS).readlines()
    assert len(ossfs.read_block(number_file, 0, 5)) == 5
    assert len(ossfs.read_block(number_file, 4, 150)) == len(NUMBERS) - 4
    assert ossfs.read_block(number_file, 20, 25) == b""

    assert ossfs.read_block(number_file, 5, None) == ossfs.read_block(
        number_file, 5, 25
    )


@pytest.mark.parametrize("size", [2**10, 2**20, 10 * 2**20])
def test_write(ossfs, test_path, size):
    file = test_path + "/test_write/file"
    data = os.urandom(size)
    with ossfs.open(file, "wb") as f:
        f.write(data)
    assert ossfs.cat(file) == data
    assert ossfs.info(file)["Size"] == len(data)
    ossfs.open(file, "wb").close()
    assert ossfs.info(file)["Size"] == 0


def test_write_fails(ossfs, test_path):
    file = test_path + "/test_write_fails/temp"
    ossfs.touch(file)
    with pytest.raises(ValueError):
        ossfs.open(file, "rb").write(b"hello")
    f = ossfs.open(file, "wb")
    f.close()
    with pytest.raises(ValueError):
        f.write(b"hello")
    with pytest.raises(FileNotFoundError):
        ossfs.open("nonexistentbucket/temp", "wb").close()


def test_write_blocks(ossfs, test_path):
    file = test_path + "/test_write_blocks/temp"
    with ossfs.open(file, "wb") as f:
        f.write(os.urandom(2 * 2**20))
        assert f.buffer.tell() == 2 * 2**20
        f.flush()
        assert f.buffer.tell() == 2 * 2**20
        f.write(os.urandom(2 * 2**20))
        f.write(os.urandom(2 * 2**20))
    assert ossfs.info(file)["Size"] == 6 * 2**20
    with ossfs.open(file, "wb", block_size=10 * 2**20) as f:
        f.write(os.urandom(15 * 2**20))
        assert f.buffer.tell() == 0
    assert ossfs.info(file)["Size"] == 15 * 2**20


def test_readline(ossfs, number_file, license_file):
    with ossfs.open("/".join([number_file]), "rb") as f_r:
        result = f_r.readline()
        expected = NUMBERS
        assert result == expected

    with ossfs.open("/".join([license_file]), "rb") as f_r, open(
        LICENSE_PATH, "rb"
    ) as f_l:
        result = f_r.readline()
        expected = f_l.readline()
        assert result == expected


def test_readline_empty(ossfs, test_path):
    file = test_path + "/test_readline_empty/empty"
    data = b""
    with ossfs.open(file, "wb") as f:
        f.write(data)
    with ossfs.open(file, "rb") as f:
        result = f.readline()
        assert result == data


def test_readline_blocksize(ossfs, test_path):
    test_file_a = test_path + "/test_readline_blocksize/a"
    data = b"ab\n" + b"a" * (10 * 2**20) + b"\nab"
    with ossfs.open(test_file_a, "wb") as f:
        f.write(data)
    with ossfs.open(test_file_a, "rb") as f:
        result = f.readline()
        expected = b"ab\n"
        assert result == expected

        result = f.readline()
        expected = b"a" * (10 * 2**20) + b"\n"
        assert result == expected

        result = f.readline()
        expected = b"ab"
        assert result == expected


def test_next(ossfs, license_file):
    with open(LICENSE_PATH, "rb") as f_l, ossfs.open(license_file) as f_r:
        expected = f_l.readline()
        result = next(f_r)
        assert result == expected


def test_iterable(ossfs, test_path):
    file = test_path + "/test_iterable/file"
    data = b"abc\n123"
    with ossfs.open(file, "wb") as f:
        f.write(data)
    with ossfs.open(file) as f, io.BytesIO(data) as g:
        for fromossfs, fromio in zip(f, g):
            assert fromossfs == fromio
        f.seek(0)
        assert f.readline() == b"abc\n"
        assert f.readline() == b"123"
        f.seek(1)
        assert f.readline() == b"bc\n"

    with ossfs.open(file) as f:
        out = list(f)
    with ossfs.open(file) as f:
        out2 = f.readlines()
    assert out == out2
    assert b"".join(out) == data


def test_file_status(ossfs, test_path):
    file = test_path + "/test_file_status/file"
    with ossfs.open(file, "wb") as f:
        assert not f.readable()
        assert not f.seekable()
        assert f.writable()

    with ossfs.open(file, "rb") as f:
        assert f.readable()
        assert f.seekable()
        assert not f.writable()


@pytest.mark.parametrize("data_size", [0, 20, 10 * 2**20])
@pytest.mark.parametrize("append_size", [0, 20, 10 * 2**20])
def test_append(ossfs, test_path, data_size, append_size):
    file = test_path + f"/test_append/file_{data_size}_{append_size}"
    data = os.urandom(data_size)
    extra = os.urandom(append_size)
    with ossfs.open(file, "wb") as f:
        f.write(data)
    assert ossfs.cat(file) == data
    with ossfs.open(file, "ab") as f:
        f.write(extra)  # append, write, small file
    assert ossfs.cat(file) == data + extra


def test_bigger_than_block_read(ossfs, number_file):
    with ossfs.open(number_file, "rb", block_size=3) as f:
        out = []
        while True:
            data = f.read(4)
            out.append(data)
            if len(data) == 0:
                break
    assert b"".join(out) == b"1234567890\n"


def test_text_io__stream_wrapper_works(ossfs, test_path):
    """Ensure using TextIOWrapper works."""
    file = test_path + "/test_text_io__stream_wrapper_works/file"
    with ossfs.open(file, "wb") as fd:
        fd.write("\u00af\\_(\u30c4)_/\u00af".encode("utf-16-le"))

    with ossfs.open(file, "rb") as fd:
        with io.TextIOWrapper(fd, "utf-16-le") as stream:
            assert stream.readline() == "\u00af\\_(\u30c4)_/\u00af"


def test_text_io__basic(ossfs, test_path):
    """Text mode is now allowed."""
    file = test_path + "/test_text_io__basic/file"
    with ossfs.open(file, "w", encoding="utf-8") as fd:
        fd.write("\u00af\\_(\u30c4)_/\u00af")

    with ossfs.open(file, "r", encoding="utf-8") as fd:
        assert fd.read() == "\u00af\\_(\u30c4)_/\u00af"


def test_text_io__override_encoding(ossfs, test_path):
    """Allow overriding the default text encoding."""
    file = test_path + "/test_text_io__override_encoding/file"

    with ossfs.open(file, "w", encoding="ibm500") as fd:
        fd.write("Hello, World!")

    with ossfs.open(file, "r", encoding="ibm500") as fd:
        assert fd.read() == "Hello, World!"


def test_readinto(ossfs, test_path):
    file = test_path + "/test_readinto/file"

    with ossfs.open(file, "wb") as fd:
        fd.write(b"Hello, World!")

    contents = bytearray(15)

    with ossfs.open(file, "rb") as fd:
        assert fd.readinto(contents) == 13

    assert contents.startswith(b"Hello, World!")


def test_seek_reads(ossfs, test_path):
    file = test_path + "/test_seek_reads/file"
    with ossfs.open(file, "wb") as f:
        f.write(os.urandom(5627146))
    with ossfs.open(file, "rb", blocksize=100) as f:
        f.seek(5561610)
        f.read(65536)

        f.seek(4)
        size = 562198
        d2 = f.read(size)
        assert len(d2) == size

        f.seek(562288)
        size = 562187
        d3 = f.read(size)
        assert len(d3) == size



"""
Test all oss object related methods
"""
# pylint:disable=invalid-name
# pylint:disable=missing-function-docstring
# pylint:disable=protected-access
# pylint:disable=consider-using-with
import os
import time

import fsspec
import pytest


def test_info(ossfs, test_path):
    "test file info"
    # test file info
    test_info_a = test_path + "/test_info/a"
    test_info_b = test_path + "/test_info/b"
    ossfs.touch(test_info_a)
    ossfs.touch(test_info_b)
    info = ossfs.info(test_info_a)
    linfo = ossfs.ls(test_info_a, detail=True)[0]
    offset = time.timezone if (time.localtime().tm_isdst == 0) else time.altzone

    assert abs(info.pop("LastModified") - linfo.pop("LastModified") + offset) <= 1
    assert info == linfo

    # test not exist dir
    new_path = test_path + "/test_info/foo"
    file_in_new_path = new_path + "/bar"
    with pytest.raises(FileNotFoundError):
        ossfs.info(new_path)

    # add a new file then we can info this new dir.
    ossfs.touch(file_in_new_path)
    ossfs.info(new_path)


def test_checksum(ossfs, test_path):
    file = test_path + "/test_check_sum/file"

    # change one file, using cache
    with ossfs.open(file, "w") as f_w:
        f_w.write("foo")
    checksum = ossfs.checksum(file)

    # Test changed contents
    with ossfs.open(file, "w") as f_w:
        f_w.write("bar")
    assert checksum != ossfs.checksum(file)
    # Test for same contents

    # sleep 1 second to make sure the last modificed changes
    time.sleep(1)
    with ossfs.open(file, "w") as f_w:
        f_w.write("foo")
    assert checksum != ossfs.checksum(file)

    # Test for nonexistent file
    ossfs.rm(file)
    with pytest.raises(FileNotFoundError):
        ossfs.checksum(file)


def test_ls_object(ossfs, test_path):
    path = test_path + "/test_ls_object/"
    assert ossfs.ls(test_path + "/nonexistent") == []
    fn = path + "accounts.1.json"
    ossfs.touch(fn)
    assert fn in ossfs.ls(path, detail=False)


def test_ls_touch(ossfs, test_path):
    path = test_path + "/test_ls_touch"
    file_a = path + "/a"
    file_b = path + "/b"
    assert not ossfs.exists(test_path + "/test_ls_touch/")
    ossfs.touch(file_a)
    ossfs.touch(file_b)
    L = ossfs.ls(test_path + "/test_ls_touch/", True)
    assert {d["Key"] for d in L} == {file_a, file_b}
    L = ossfs.ls(test_path + "/test_ls_touch/", False)
    assert set(L) == {file_a, file_b}


def test_isfile(ossfs, test_path):
    "test isfile in ossfs"
    file_a = test_path + "/test_isfile/a"
    assert not ossfs.isfile("")
    assert not ossfs.isfile("/")
    assert not ossfs.isfile(test_path)

    assert not ossfs.isfile(file_a)
    ossfs.touch(file_a)
    assert ossfs.isfile(file_a)


def test_isdir(ossfs, test_path):
    "test isdir in ossfs"
    file_a = test_path + "/test_isdir/a"
    ossfs.touch(file_a)
    assert ossfs.isdir("")
    assert ossfs.isdir("/")
    assert not ossfs.isdir(file_a)
    assert ossfs.isdir(test_path)
    assert ossfs.isdir(test_path + "/test_isdir/")
    assert ossfs.isdir(test_path + "/test_isdir")


def test_rm(ossfs, test_path):
    file_a = test_path + "/test_rm/a"
    nest_file = test_path + "/test_rm/nested/nested2/file1"
    ossfs.touch(file_a)
    assert ossfs.exists(file_a)
    ossfs.rm(file_a)
    assert not ossfs.exists(file_a)
    with pytest.raises(FileNotFoundError):
        ossfs.rm("nonexistent")
    ossfs.touch(nest_file)
    ossfs.rm(test_path + "/test_rm/nested", recursive=True)
    assert not ossfs.exists(nest_file)


def test_bulk_delete(ossfs, test_path):
    nest_file1 = test_path + "/test_bulk_delete/nested/nested2/file1"
    nest_file2 = test_path + "/test_bulk_delete/nested/nested2/file2"
    ossfs.touch(nest_file1)
    ossfs.touch(nest_file2)
    filelist = ossfs.find(test_path + "/test_bulk_delete")
    ossfs.rm(filelist)
    assert not ossfs.exists(test_path + "/test_bulk_delete/nested/nested2/file1")


def test_ossfs_file_access(ossfs, number_file):
    data = b"1234567890\n"
    assert ossfs.cat(number_file) == data
    assert ossfs.head(number_file, 3) == data[:3]
    assert ossfs.tail(number_file, 3) == data[-3:]
    assert ossfs.tail(number_file, 10000) == data
    assert ossfs.info(number_file)["Size"] == len(data)


def test_du(ossfs, test_path):
    path = test_path + "/test_du"
    file1 = path + "/file1"
    file2 = path + "/dir/file2"
    file3 = path + "/dir/file3"
    with ossfs.open(file1, "wb") as f:
        f.write(b"1234567890")
    with ossfs.open(file2, "wb") as f:
        f.write(b"12345")
    with ossfs.open(file3, "wb") as f:
        f.write(b"1234567890" * 2)
    d = ossfs.du(path, total=False)
    assert all(isinstance(v, int) and v >= 0 for v in d.values())
    assert d[file1] == 10
    assert d[file2] == 5
    assert d[file3] == 20

    assert ossfs.du(path, total=True) == 35


def test_ossfs_ls(ossfs, test_path):
    path = test_path + "/test_ossfs_ls"
    fn = path + "/nested/file1"
    ossfs.touch(fn)
    assert fn not in ossfs.ls(path, detail=False)
    assert fn in ossfs.ls(path + "/nested/", detail=False)
    assert fn in ossfs.ls(path + "/nested", detail=False)
    L = ossfs.ls(path + "/nested", detail=True)
    assert all(isinstance(item, dict) for item in L)


def test_ossfs_big_ls(ossfs, test_path):
    path = test_path + "/test_ossfs_big_ls"
    if not ossfs.exists(path):
        for x in range(120):
            ossfs.touch(path + f"/{x}.part")

    assert len(ossfs.find(path, connect_timeout=600)) == 120


def test_ossfs_glob(ossfs, test_path):
    path = test_path + "/test_ossfs_glob"
    fn = path + "/nested/file.dat"
    fn2 = path + "/nested/filedat"
    ossfs.touch(fn)
    ossfs.touch(fn2)
    assert fn not in ossfs.glob(path + "/")
    assert fn not in ossfs.glob(path + "/*")
    assert fn not in ossfs.glob(path + "/nested")
    assert fn in ossfs.glob(path + "/nested/*")
    assert fn in ossfs.glob(path + "/nested/file*")
    assert all(
        any(p.startswith(f + "/") or p == f for p in ossfs.find(path))
        for f in ossfs.glob(path + "/nested/*")
    )
    out = ossfs.glob(path + "/nested/*")
    assert {fn, fn2} == set(out)

    # Make sure glob() deals with the dot character (.) correctly.
    assert fn in ossfs.glob(path + "/nested/file.*")
    assert fn2 not in ossfs.glob(path + "/nested/file.*")


def test_copy(ossfs, number_file, test_path):
    new_file = test_path + "/test_copy/file"
    ossfs.copy(number_file, new_file)
    assert ossfs.cat(number_file) == ossfs.cat(new_file)


def test_move(ossfs, number_file, test_path):
    from_file = test_path + "/test_move/from"
    to_file = test_path + "/test_move/to"
    ossfs.copy(number_file, from_file)
    assert ossfs.exists(from_file)
    assert not ossfs.exists(to_file)
    data = ossfs.cat(from_file)
    ossfs.mv(from_file, to_file)
    assert ossfs.cat(to_file) == data
    assert not ossfs.exists(from_file)


@pytest.mark.parametrize("size", [2**10, 2**20, 10 * 2**20])
def test_get_put(ossfs, tmpdir, test_path, size):
    local_file = str(tmpdir.join("number"))
    data = os.urandom(size)
    open(local_file, "wb").write(data)

    remote_file = test_path + "/test_get_put/file"
    ossfs.put(local_file, remote_file)
    assert ossfs.exists(remote_file)
    assert ossfs.cat(remote_file) == data

    get_file = str(tmpdir.join("get"))
    ossfs.get(remote_file, get_file)
    assert open(get_file, "rb").read() == data


@pytest.mark.parametrize("size", [2**10, 2**20, 10 * 2**20])
def test_pipe_cat_big(ossfs, size, test_path):
    bigfile = test_path + "/test_get_put_big/bigfile"
    data = b"1234567890A" * size
    ossfs.pipe(bigfile, data)
    assert ossfs.cat(bigfile) == data


def test_errors(ossfs, test_path):
    with pytest.raises(FileNotFoundError):
        ossfs.open(test_path + "/test_errors/none", "rb")

    with pytest.raises(FileNotFoundError):
        ossfs.mv(test_path + "/tmp/test/shfoshf/x", "tmp/test/shfoshf/y")

    with pytest.raises(ValueError):
        ossfs.open("x", "rb")

    with pytest.raises(ValueError):
        ossfs.rm("/non_exist_bucket")

    with pytest.raises(ValueError):
        with ossfs.open(test_path + "/temp", "wb") as f:
            f.read()

    with pytest.raises(ValueError):
        f = ossfs.open(test_path + "/temp", "rb")
        f.close()
        f.read()


def test_touch(ossfs, test_path):
    # create
    fn = test_path + "/test_touch/file"
    assert not ossfs.exists(fn)
    ossfs.touch(fn)
    assert ossfs.exists(fn)
    assert ossfs.size(fn) == 0

    # truncates
    with ossfs.open(fn, "wb") as f:
        f.write(b"data")
    assert ossfs.size(fn) == 4

    # exists error
    with ossfs.open(fn, "wb") as f:
        f.write(b"data")
    assert ossfs.size(fn) == 4
    ossfs.touch(fn, truncate=False)
    assert ossfs.size(fn) == 4


def test_cat_missing(ossfs, test_path):
    fn0 = test_path + "/test_cat_missing/file0"
    fn1 = test_path + "/test_cat_missing/file1"
    ossfs.touch(fn0)
    with pytest.raises(FileNotFoundError):
        ossfs.cat([fn0, fn1], on_error="raise")
    out = ossfs.cat([fn0, fn1], on_error="omit")
    assert list(out) == [fn0]
    out = ossfs.cat([fn0, fn1], on_error="return")
    assert fn1 in out
    assert isinstance(out[fn1], FileNotFoundError)


def test_get_directories(ossfs, tmpdir, test_path):
    path = test_path + "/test_get_directories"
    ossfs.touch(path + "/dir/dirkey/key0")
    ossfs.touch(path + "/dir/dirkey/key1")
    ossfs.touch(path + "/dir/dir/key")
    d = str(tmpdir)
    ossfs.get(path + "/dir/", d, recursive=True)
    assert {"dirkey", "dir"} == set(os.listdir(d))
    assert ["key"] == os.listdir(os.path.join(d, "dir"))
    assert {"key0", "key1"} == set(os.listdir(os.path.join(d, "dirkey")))


def test_lsdir(ossfs, test_path):
    path = test_path + "/test_lsdir/"
    file = path + "file"
    ossfs.touch(file)
    assert file in ossfs.ls(path, detail=False)


def test_modified(ossfs, test_path):
    dir_path = test_path + "/test_modified/"
    file_path = dir_path + "file"
    notexist_path = dir_path + "notexist"

    # Test file
    ossfs.touch(file_path)
    modified = ossfs.modified(path=file_path)
    assert isinstance(modified, int)

    # Test file
    with pytest.raises(FileNotFoundError):
        modified = ossfs.modified(path=notexist_path)

    # Test directory
    with pytest.raises(NotImplementedError):
        modified = ossfs.modified(path=dir_path)

    # Test bucket
    with pytest.raises(NotImplementedError):
        ossfs.modified(path=test_path)


def test_get_file_info_with_selector(ossfs, test_path):
    base_dir = test_path + "/test_get_file_info_with_selector"
    file_a = base_dir + "/test_file_a"
    file_b = base_dir + "/test_file_b"
    dir_a = base_dir + "/test_dir_a"
    file_c = base_dir + "/test_dir_a/test_file_c"

    ossfs.touch(file_a)
    ossfs.touch(file_b)
    ossfs.touch(file_c)

    infos = ossfs.find(base_dir, maxdepth=None, withdirs=True, detail=True)
    assert len(infos) == 4

    for info in infos.values():
        if info["name"].endswith(file_a):
            assert info["type"] == "file"
        elif info["name"].endswith(file_b):
            assert info["type"] == "file"
        elif info["name"].endswith(file_c):
            assert info["type"] == "file"
        elif info["name"].rstrip("/").endswith(dir_a):
            assert info["type"] == "directory"
        else:
            raise ValueError(f"unexpected path {info['name']}")


def test_same_name_but_no_exact(ossfs, test_path):
    path = test_path + "/test_same_name_but_no_exact"
    ossfs.touch(path + "/very/similar/prefix1")
    ossfs.touch(path + "/very/similar/prefix2")
    ossfs.touch(path + "/very/similar/prefix3/something")
    assert not ossfs.exists(path + "/very/similar/prefix")
    assert not ossfs.exists(path + "/very/similar/prefi")
    assert not ossfs.exists(path + "/very/similar/pref")

    assert ossfs.exists(path + "/very/similar/")
    assert ossfs.exists(path + "/very/similar/prefix1")
    assert ossfs.exists(path + "/very/similar/prefix2")
    assert ossfs.exists(path + "/very/similar/prefix3")
    assert ossfs.exists(path + "/very/similar/prefix3/")
    assert ossfs.exists(path + "/very/similar/prefix3/something")

    assert not ossfs.exists(path + "/very/similar/prefix3/some")

    ossfs.touch(path + "/starting/very/similar/prefix")

    assert not ossfs.exists(path + "/starting/very/similar/prefix1")
    assert not ossfs.exists(path + "/starting/very/similar/prefix3")
    assert not ossfs.exists(path + "/starting/very/similar/prefix3/")
    assert not ossfs.exists(path + "/starting/very/similar/prefix3/something")

    assert ossfs.exists(path + "/starting/very/similar/prefix")
    assert not ossfs.exists(path + "/starting/very/similar/prefix/")


def test_leading_forward_slash(ossfs, test_path):
    path = test_path + "/test_leading_forward_slash"
    ossfs.touch(path + "/some/file")
    assert ossfs.ls(path + "/some/")
    path = path.lstrip("/")
    assert ossfs.exists(path + "/some/file")
    assert ossfs.exists("/" + path + "/some/file")


def test_find_with_prefix(ossfs, test_path):
    path = test_path + "/test_find_with_prefix"
    if not ossfs.exists(path):
        for cursor in range(100):
            ossfs.touch(path + f"/prefixes/test_{cursor}")
        ossfs.touch(path + "/prefixes2")

    assert len(ossfs.find(path + "/prefixes")) == 100
    assert len(ossfs.find(path, prefix="prefixes")) == 101

    assert len(ossfs.find(path + "/prefixes/test_")) == 0
    assert len(ossfs.find(path + "/prefixes", prefix="test_")) == 100
    assert len(ossfs.find(path + "/prefixes/", prefix="test_")) == 100

    test_1s = ossfs.find(path + "/prefixes/test_1")
    assert len(test_1s) == 1
    assert test_1s[0] == path + "/prefixes/test_1"

    test_1s = ossfs.find(path + "/prefixes/", prefix="test_1")
    assert len(test_1s) == 11
    assert test_1s == [path + "/prefixes/test_1"] + [
        path + f"/prefixes/test_{cursor}" for cursor in range(10, 20)
    ]


WRITE_BLOCK_SIZE = 2**13  # 8KB blocks
READ_BLOCK_SIZE = 2**14  # 16KB blocks


def test_get_put_file(ossfs, tmpdir, test_path):
    src_file = str(tmpdir / "source")
    src2_file = str(tmpdir / "source_2")
    dest_file = test_path + "/get_put_file/dest"

    data = b"test" * 2**20

    # pylint: disable=missing-class-docstring
    class EventLogger(fsspec.Callback):
        def __init__(self):
            self.events = []
            super().__init__()

        def set_size(self, size):
            self.events.append(("set_size", size))

        def absolute_update(self, value):
            self.events.append(("absolute_update", value))

    with open(src_file, "wb") as stream:
        stream.write(data)

    event_logger = EventLogger()
    ossfs.put_file(src_file, dest_file, callback=event_logger)
    assert ossfs.exists(dest_file)

    assert event_logger.events[0] == ("set_size", len(data))
    assert len(event_logger.events[1:]) == len(data) // WRITE_BLOCK_SIZE

    event_logger = EventLogger()
    ossfs.get_file(dest_file, src2_file, callback=event_logger)
    with open(src2_file, "rb") as stream:
        assert stream.read() == data

    assert event_logger.events[0] == ("set_size", len(data))
    assert len(event_logger.events[1:]) == len(data) // READ_BLOCK_SIZE



"""
Pytest setup
"""
# pylint: disable=missing-function-docstring
# pylint: disable=redefined-outer-name
import os
import pathlib
import subprocess
import uuid

import oss2
import pytest
import requests

from ossfs import OSSFileSystem

PORT = 5555
AccessKeyId = os.environ.get("OSS_ACCESS_KEY_ID", "")
AccessKeySecret = os.environ.get("OSS_SECRET_ACCESS_KEY", "")
LICENSE_PATH = os.path.join(pathlib.Path(__file__).parent.parent.resolve(), "LICENSE")
NUMBERS = b"1234567890\n"


test_id = uuid.uuid4()


@pytest.fixture(scope="session")
def emulator_endpoint():
    return f"http://127.0.0.1:{PORT}/"


@pytest.fixture(scope="session")
def endpoint():
    return os.environ.get("OSS_ENDPOINT")


@pytest.fixture(scope="session")
def test_bucket_name():
    return os.environ.get("OSS_TEST_BUCKET_NAME")


@pytest.fixture(scope="session")
def test_directory():
    return f"ossfs_test/{test_id}"


@pytest.fixture(scope="session")
def test_path(test_bucket_name, test_directory):
    return f"/{test_bucket_name}/{test_directory}"


@pytest.fixture()
def oss_emulator_server_start(emulator_endpoint):
    """
    Start a local emulator server
    """
    with subprocess.Popen(f"ruby bin/emulator -r store -p {PORT}"):
        try:
            result = requests.get(emulator_endpoint, timeout=5)
            if result.ok:
                yield
        except TimeoutError as err:
            raise Exception("emulator start timeout") from err


@pytest.fixture(scope="session")
def init_config(endpoint):
    result = {}
    result["key"] = AccessKeyId
    result["secret"] = AccessKeySecret
    result["endpoint"] = endpoint
    return result


@pytest.fixture()
def ossfs(init_config):
    return OSSFileSystem(**init_config)


def put_object(endpoint, bucket_name, filename, contents):
    auth = oss2.Auth(AccessKeyId, AccessKeySecret)
    bucket = oss2.Bucket(auth, endpoint, bucket_name)
    bucket.put_object(filename, contents)


def put_file(endpoint, bucket_name, key, filename):
    auth = oss2.Auth(AccessKeyId, AccessKeySecret)
    bucket = oss2.Bucket(auth, endpoint, bucket_name)
    bucket.put_object_from_file(key, filename)


@pytest.fixture(scope="session")
def file_in_anonymous(endpoint, test_directory, test_bucket_name):
    bucket = f"{test_bucket_name}-anonymous"
    file = f"{test_directory}/file"
    put_object(endpoint, bucket, file, "foobar")
    return f"/{bucket}/{file}"


@pytest.fixture(scope="session")
def number_file(test_bucket_name, endpoint, test_directory):
    filename = f"{test_directory}/number"
    put_object(endpoint, test_bucket_name, filename, NUMBERS)
    return f"/{test_bucket_name}/{filename}"


@pytest.fixture(scope="session")
def license_file(test_bucket_name, endpoint, test_directory):
    filename = f"{test_directory}/LICENSE"
    put_file(endpoint, test_bucket_name, filename, LICENSE_PATH)
    return f"/{test_bucket_name}/{filename}"



"""
OSSFS
----------------------------------------------------------------
A pythonic file-systems interface to OSS (Object Storage Service)
"""
from .core import OSSFile, OSSFileSystem

__all__ = ["OSSFile", "OSSFileSystem"]



"""
Three different login test
Sts, anonymous, accesskey
"""
# pylint:disable=invalid-name
# pylint:disable=missing-function-docstring
import json
import os

import oss2
import pytest
from aliyunsdkcore import client
from aliyunsdksts.request.v20150401 import AssumeRoleRequest

from ossfs import OSSFileSystem

STSAccessKeyId = os.getenv("OSS_TEST_STS_ID", "")
STSAccessKeySecret = os.getenv("OSS_TEST_STS_KEY", "")
STSArn = os.getenv("OSS_TEST_STS_ARN", "")


def fetch_sts_token(access_key_id, access_key_secret, role_arn):
    clt = client.AcsClient(access_key_id, access_key_secret, "cn-hangzhou")
    req = AssumeRoleRequest.AssumeRoleRequest()

    req.set_accept_format("json")
    req.set_RoleArn(role_arn)
    req.set_RoleSessionName("oss-python-sdk-example")

    body = clt.do_action_with_exception(req)

    j = json.loads(oss2.to_unicode(body))

    access_key_id = j["Credentials"]["AccessKeyId"]
    access_key_secret = j["Credentials"]["AccessKeySecret"]
    security_token = j["Credentials"]["SecurityToken"]

    return access_key_id, access_key_secret, security_token


def test_access_key_login(ossfs, test_bucket_name):
    ossfs.ls(test_bucket_name)


def test_sts_login(endpoint, test_bucket_name):
    key, secret, token = fetch_sts_token(STSAccessKeyId, STSAccessKeySecret, STSArn)
    ossfs = OSSFileSystem(
        key=key,
        secret=secret,
        token=token,
        endpoint=endpoint,
    )
    ossfs.ls(test_bucket_name)


def test_set_endpoint(endpoint, test_bucket_name, monkeypatch):
    key, secret, token = fetch_sts_token(STSAccessKeyId, STSAccessKeySecret, STSArn)
    monkeypatch.delenv("OSS_ENDPOINT")
    ossfs = OSSFileSystem(key=key, secret=secret, token=token, endpoint=None)
    with pytest.raises(ValueError):
        ossfs.ls(test_bucket_name)
    ossfs.set_endpoint(endpoint)
    ossfs.ls(test_bucket_name)


def test_env_endpoint(endpoint, test_bucket_name, monkeypatch):
    key, secret, token = fetch_sts_token(STSAccessKeyId, STSAccessKeySecret, STSArn)
    monkeypatch.setenv("OSS_ENDPOINT", endpoint)
    ossfs = OSSFileSystem(key=key, secret=secret, token=token, endpoint=None)
    ossfs.ls(test_bucket_name)


def test_anonymous_login(file_in_anonymous, endpoint):
    ossfs = OSSFileSystem(endpoint=endpoint)
    ossfs.cat(f"{file_in_anonymous}")



"""
Code of OSSFileSystem and OSSFile
"""
import copy
import logging
import os
import re
from datetime import datetime
from hashlib import sha256
from typing import Dict, List, Optional, Tuple, Union

import oss2
from fsspec.spec import AbstractBufferedFile, AbstractFileSystem
from fsspec.utils import stringify_path

from ossfs.exceptions import translate_boto_error

logger = logging.getLogger("ossfs")
logging.getLogger("oss2").setLevel(logging.CRITICAL)
logging.getLogger("urllib3").setLevel(logging.WARNING)


def _as_progress_handler(callback):
    if callback is None:
        return None

    sent_total = False

    def progress_handler(absolute_progress, total_size):
        nonlocal sent_total
        if not sent_total:
            callback.set_size(total_size)
            sent_total = True

        callback.absolute_update(absolute_progress)

    return progress_handler


class OSSFileSystem(AbstractFileSystem):  # pylint:disable=too-many-public-methods
    # pylint:disable=no-value-for-parameter
    """
    A pythonic file-systems interface to OSS (Object Storage Service)
    """

    protocol = "oss"
    SIMPLE_TRANSFER_THRESHOLD = oss2.defaults.multiget_threshold

    def __init__(
        self,
        endpoint: str,
        key: Optional[str] = None,
        secret: Optional[str] = None,
        token: Optional[str] = None,
        default_cache_type: Optional[str] = "readahead",
        **kwargs,  # pylint: disable=too-many-arguments
    ):
        """
        Parameters
        ----------
        key : string (None)
            If not anonymous, use this access key ID, if specified
        secret : string (None)
            If not anonymous, use this secret access key, if specified
        token : string (None)
            If not anonymous, use this security token, if specified
        endpoint : string (None)
            Default endpoints of the fs
            Endpoints are the adderss where OSS locate
            like: http://oss-cn-hangzhou.aliyuncs.com or
                        https://oss-me-east-1.aliyuncs.com
        """
        super().__init__(**kwargs)
        if token:
            self._auth = oss2.StsAuth(key, secret, token)
        elif key:
            self._auth = oss2.Auth(key, secret)
        else:
            self._auth = oss2.AnonymousAuth()
        self._endpoint = endpoint or os.getenv("OSS_ENDPOINT")
        if self._endpoint is None:
            logger.warning(
                "OSS endpoint is not set, OSSFS could not work properly"
                "without a endpoint, please set it manually with "
                "`ossfs.set_endpoint` later"
            )
        self._default_cache_type = default_cache_type
        self._session = oss2.Session()

    def set_endpoint(self, endpoint: str):
        """
        Reset the endpoint for ossfs
        endpoint : string (None)
            Default endpoints of the fs
            Endpoints are the adderss where OSS locate
            like: http://oss-cn-hangzhou.aliyuncs.com or
        """
        if not endpoint:
            raise ValueError("Not a valid endpoint")
        self._endpoint = endpoint

    def _get_bucket(
        self, bucket_name: str, connect_timeout: Optional[int] = None
    ) -> oss2.Bucket:
        """
        get the new bucket instance
        """
        if not self._endpoint:
            raise ValueError("endpoint is required")
        try:
            return oss2.Bucket(
                self._auth,
                self._endpoint,
                bucket_name,
                session=self._session,
                connect_timeout=connect_timeout,
                app_name="ossfs",
            )
        except oss2.exceptions.ClientError as err:
            raise ValueError(bucket_name) from err

    def _call_oss(
        self,
        method_name: str,
        *args,
        bucket: Optional[str] = None,
        timeout: Optional[int] = None,
        retry: int = 3,
        **kwargs,
    ):
        if bucket:
            service = self._get_bucket(bucket, timeout)
        else:
            service = oss2.Service(
                self._auth,
                endpoint=self._endpoint,
                connect_timeout=timeout,
            )
        for count in range(retry):
            try:
                method = getattr(service, method_name, None)
                if not method:
                    method = getattr(oss2, method_name)
                    logger.debug("CALL: %s - %s - %s", method.__name__, args, kwargs)
                    out = method(service, *args, **kwargs)
                else:
                    logger.debug("CALL: %s - %s - %s", method.__name__, args, kwargs)
                    out = method(*args, **kwargs)
                return out
            except oss2.exceptions.RequestError as err:
                logger.debug("Retryable error: %s, try %s times", err, count + 1)
                error = err
            except oss2.exceptions.OssError as err:
                logger.debug("Nonretryable error: %s", err)
                error = err
                break
        raise translate_boto_error(error)

    def split_path(self, path: str) -> Tuple[str, str]:
        """
        Normalise object path string into bucket and key.
        Parameters
        ----------
        path : string
            Input path, like `/mybucket/path/to/file`
        Examples
        --------
        >>> split_path("/mybucket/path/to/file")
        ['mybucket', 'path/to/file' ]
        >>> split_path("
        /mybucket/path/to/versioned_file?versionId=some_version_id
        ")
        ['mybucket', 'path/to/versioned_file', 'some_version_id']
        """
        path = self._strip_protocol(path)
        path = path.lstrip("/")
        if "/" not in path:
            return path, ""
        bucket_name, obj_name = path.split("/", 1)
        return bucket_name, obj_name

    def _open(
        self,
        path,
        mode="rb",
        block_size=None,
        autocommit=True,
        cache_options=None,
        **kwargs,  # pylint: disable=too-many-arguments
    ):
        """
        Open a file for reading or writing.
        Parameters
        ----------
        path: str
            File location
        mode: str
            'rb', 'wb', etc.
        autocommit: bool
            If False, writes to temporary file that only gets put in final
            location upon commit
        kwargs
        Returns
        -------
        OSSFile instance
        """
        cache_type = kwargs.pop("cache_type", self._default_cache_type)
        return OSSFile(
            self,
            path,
            mode,
            block_size,
            autocommit,
            cache_options=cache_options,
            cache_type=cache_type,
            **kwargs,
        )

    @classmethod
    def _strip_protocol(cls, path: Union[str, List[str]]):
        """Turn path from fully-qualified to file-system-specifi
        Parameters
        ----------
        path : string
            Input path, like
            `http://oss-cn-hangzhou.aliyuncs.com/mybucket/myobject`
            `oss://mybucket/myobject`
        Examples
        --------
        >>> _strip_protocol(
            "http://oss-cn-hangzhou.aliyuncs.com/mybucket/myobject"
            )
        ('/mybucket/myobject')
        >>> _strip_protocol(
            "oss://mybucket/myobject"
            )
        ('/mybucket/myobject')
        """
        if isinstance(path, list):
            return [cls._strip_protocol(p) for p in path]
        path_string: str = stringify_path(path)
        if path_string.startswith("oss://"):
            path_string = path_string[5:]

        parser_re = r"https?://(?P<endpoint>oss.+aliyuncs\.com)(?P<path>/.+)"
        matcher = re.compile(parser_re).match(path_string)
        if matcher:
            path_string = matcher["path"]
        return path_string or cls.root_marker

    def _ls_bucket(self, connect_timeout) -> List[Dict]:
        result = []
        for bucket in self._call_oss("BucketIterator", timeout=connect_timeout):
            result.append(
                {
                    "name": bucket.name,
                    "Key": bucket.name,
                    "type": "directory",
                    "size": 0,
                    "Size": 0,
                    "StorageClass": "BUCKET",
                    "CreateTime": bucket.creation_date,
                }
            )
        return result

    def _ls_object(self, path: str, connect_timeout) -> List[Dict]:
        bucket_name, obj_name = self.split_path(path)
        if not obj_name or not bucket_name:
            return []
        if not self._call_oss(
            "object_exists",
            obj_name,
            bucket=bucket_name,
            timeout=connect_timeout,
        ):
            return []
        simplifiedmeta = self._call_oss(
            "get_object_meta",
            obj_name,
            bucket=bucket_name,
            timeout=connect_timeout,
        )
        info = {
            "name": path,
            "Key": path,
            "type": "file",
            "size": int(simplifiedmeta.headers["Content-Length"]),
            "Size": int(simplifiedmeta.headers["Content-Length"]),
            "StorageClass": "OBJECT",
        }
        if "Last-Modified" in simplifiedmeta.headers:
            info["LastModified"] = int(
                datetime.strptime(
                    simplifiedmeta.headers["Last-Modified"],
                    "%a, %d %b %Y %H:%M:%S %Z",
                ).timestamp()
            )
        return [info]

    def _get_object_info_list(
        self,
        bucket_name: str,
        prefix: str,
        delimiter: str,
        connect_timeout: Optional[int],
    ):
        """
        Wrap oss2.ObjectIterator return values into a
        fsspec form of file info
        """
        result = []
        for obj in self._call_oss(
            "ObjectIterator",
            prefix=prefix,
            delimiter=delimiter,
            bucket=bucket_name,
            timeout=connect_timeout,
        ):
            data = {
                "name": f"{bucket_name}/{obj.key}",
                "Key": f"{bucket_name}/{obj.key}",
                "type": "file",
                "size": obj.size,
                "Size": obj.size,
                "StorageClass": "OBJECT",
            }
            if obj.last_modified:
                data["LastModified"] = obj.last_modified
            if obj.is_prefix():
                data["type"] = "directory"
                data["size"] = 0
                data["Size"] = 0
            result.append(data)
        return result

    def _ls_dir(
        self,
        path: str,
        delimiter: str = "/",
        prefix: Optional[str] = None,
        connect_timeout: int = None,
    ) -> List[Dict]:
        norm_path = path.strip("/")
        bucket_name, key = self.split_path(norm_path)
        if not prefix:
            prefix = ""

        if not delimiter or prefix:
            if key:
                prefix = f"{key}/{prefix}"
            infos = self._get_object_info_list(
                bucket_name, prefix, delimiter, connect_timeout
            )
        else:
            if key:
                prefix = f"{key}/"
            if norm_path not in self.dircache:
                self.dircache[norm_path] = self._get_object_info_list(
                    bucket_name, prefix, delimiter, connect_timeout
                )
            infos = copy.deepcopy(self.dircache[norm_path])
        if path.startswith("/"):
            for info in infos:
                info["name"] = f'/{info["name"]}'
                info["Key"] = f'/{info["Key"]}'
        return infos

    def ls(self, path, detail=True, **kwargs):
        connect_timeout = kwargs.pop("connect_timeout", 60)
        bucket_name, _ = self.split_path(path)
        if bucket_name:
            try:
                infos = self._ls_dir(path, connect_timeout=connect_timeout)
            except oss2.exceptions.AccessDenied:
                infos = []
            if not infos:
                infos = self._ls_object(path, connect_timeout)
        else:
            infos = self._ls_bucket(connect_timeout)

        if not infos:
            return infos
        if detail:
            return sorted(infos, key=lambda i: i["name"])
        return sorted(info["name"] for info in infos)

    def find(self, path, maxdepth=None, withdirs=False, detail=False, **kwargs):
        """List all files below path.

        Like posix ``find`` command without conditions

        Parameters
        ----------
        path : str
        maxdepth: int or None
            If not None, the maximum number of levels to descend
        withdirs: bool
            Whether to include directory paths in the output. This is True
            when used by glob, but users usually only want files.
        kwargs are passed to ``ls``.
        """
        path = self._strip_protocol(path)
        out = {}
        prefix = kwargs.pop("prefix", None)
        if (withdirs or maxdepth) and prefix:
            raise ValueError(
                "Can not specify 'prefix' option alongside "
                "'withdirs'/'maxdepth' options."
            )
        if prefix:
            connect_timeout = kwargs.get("connect_timeout", None)
            for info in self._ls_dir(
                path,
                delimiter="",
                prefix=prefix,
                connect_timeout=connect_timeout,
            ):
                out.update({info["name"]: info})
        else:
            for _, dirs, files in self.walk(path, maxdepth, detail=True, **kwargs):
                if withdirs:
                    files.update(dirs)
                out.update({info["name"]: info for name, info in files.items()})
            if self.isfile(path) and path not in out:
                # walk works on directories, but find should also return [path]
                # when path happens to be a file
                out[path] = {}
        names = sorted(out)
        if not detail:
            return names
        return {name: out[name] for name in names}

    def _directory_exists(self, dirname: str, **kwargs):
        connect_timeout = kwargs.pop("connect_timeout", None)
        ls_result = self._ls_dir(dirname, connect_timeout=connect_timeout)
        return bool(ls_result)

    def _bucket_exist(self, bucket_name):
        if not bucket_name:
            return False
        try:
            self._call_oss("get_bucket_info", bucket=bucket_name)
        except oss2.exceptions.OssError:
            return False
        return True

    def exists(self, path, **kwargs):
        """Is there a file at the given path"""
        bucket_name, obj_name = self.split_path(path)

        if not self._bucket_exist(bucket_name):
            return False

        connect_timeout = kwargs.get("connect_timeout", None)
        if not obj_name:
            return True

        if self._call_oss(
            "object_exists",
            obj_name,
            bucket=bucket_name,
            timeout=connect_timeout,
        ):
            return True

        return self._directory_exists(path, **kwargs)

    def ukey(self, path):
        """Hash of file properties, to tell if it has changed"""
        bucket_name, obj_name = self.split_path(path)
        obj_stream = self._call_oss("get_object", obj_name, bucket=bucket_name)
        return obj_stream.server_crc

    def checksum(self, path):
        """Unique value for current version of file

        If the checksum is the same from one moment to another, the contents
        are guaranteed to be the same. If the checksum changes, the contents
        *might* have changed.

        This should normally be overridden; default will probably capture
        creation/modification timestamp (which would be good) or maybe
        access timestamp (which would be bad)
        """
        return sha256(
            (str(self.ukey(path)) + str(self.info(path))).encode()
        ).hexdigest()

    def cp_file(self, path1, path2, **kwargs):
        """
        Copy within two locations in the filesystem
        # todo: big file optimization
        """
        bucket_name1, obj_name1 = self.split_path(path1)
        bucket_name2, obj_name2 = self.split_path(path2)
        if bucket_name1 != bucket_name2:
            tempdir = "." + self.ukey(path1)
            self.get_file(path1, tempdir, **kwargs)
            self.put_file(tempdir, path2, **kwargs)
            os.remove(tempdir)
        else:
            connect_timeout = kwargs.pop("connect_timeout", None)
            self._call_oss(
                "copy_object",
                bucket_name1,
                obj_name1,
                obj_name2,
                bucket=bucket_name1,
                timeout=connect_timeout,
            )
        self.invalidate_cache(self._parent(path2))

    def _rm(self, path: Union[str, List[str]]):
        """Delete files.

        Parameters
        ----------
        path: str or list of str
            File(s) to delete.
        """
        if isinstance(path, list):
            for file in path:
                self._rm(file)
            return
        bucket_name, obj_name = self.split_path(path)
        self._call_oss("delete_object", obj_name, bucket=bucket_name)
        self.invalidate_cache(self._parent(path))

    def rm(self, path: Union[str, List[str]], recursive=False, maxdepth=None):
        """Delete files.

        Parameters
        ----------
        path: str or list of str
            File(s) to delete.
        recursive: bool
            If file(s) are directories, recursively delete contents and then
            also remove the directory
        maxdepth: int or None
            Depth to pass to walk for finding files to delete, if recursive.
            If None, there will be no limit and infinite recursion may be
            possible.
        """

        if isinstance(path, list):
            for file in path:
                self.rm(file)
            return

        bucket_name, _ = self.split_path(path)
        path_expand = self.expand_path(path, recursive=recursive, maxdepth=maxdepth)
        path_expand = [self.split_path(file)[1] for file in path_expand]

        def chunks(lst: list, num: int):
            for i in range(0, len(lst), num):
                yield lst[i : i + num]

        for files in chunks(path_expand, 1000):
            self._call_oss("batch_delete_objects", files, bucket=bucket_name)

        self.invalidate_cache(self._parent(path))

    def get_path(self, rpath, lpath, **kwargs):
        """
        Copy single remote path to local
        """
        if self.isdir(rpath):
            os.makedirs(lpath, exist_ok=True)
        else:
            self.get_file(rpath, lpath, **kwargs)

    def get_file(
        self, rpath, lpath, callback=None, **kwargs
    ):  # pylint: disable=arguments-differ
        """
        Copy single remote file to local
        """
        kwargs.setdefault("progress_callback", _as_progress_handler(callback))
        if self.isdir(rpath):
            os.makedirs(lpath, exist_ok=True)
        else:
            bucket_name, obj_name = self.split_path(rpath)
            connect_timeout = kwargs.pop("connect_timeout", None)
            bucket = self._get_bucket(bucket_name, connect_timeout)
            if self.size(rpath) >= self.SIMPLE_TRANSFER_THRESHOLD:
                oss2.resumable_download(bucket, obj_name, lpath, **kwargs)
            else:
                self._call_oss(
                    "get_object_to_file",
                    obj_name,
                    lpath,
                    bucket=bucket_name,
                    timeout=connect_timeout,
                    **kwargs,
                )

    def put_file(
        self, lpath, rpath, callback=None, **kwargs
    ):  # pylint: disable=arguments-differ
        """
        Copy single file to remote
        """
        kwargs.setdefault("progress_callback", _as_progress_handler(callback))
        if os.path.isdir(lpath):
            self.makedirs(rpath, exist_ok=True)
        else:
            bucket_name, obj_name = self.split_path(rpath)
            connect_timeout = kwargs.pop("connect_timeout", None)
            bucket = self._get_bucket(bucket_name, connect_timeout)
            if os.path.getsize(lpath) >= self.SIMPLE_TRANSFER_THRESHOLD:
                oss2.resumable_upload(bucket, obj_name, lpath, **kwargs)
            else:
                self._call_oss(
                    "put_object_from_file",
                    obj_name,
                    lpath,
                    bucket=bucket_name,
                    timeout=connect_timeout,
                    **kwargs,
                )
        self.invalidate_cache(self._parent(rpath))

    def created(self, path):
        """Return the created timestamp of a file as a datetime.datetime"""
        bucket_name, obj_name = self.split_path(path)
        if obj_name:
            raise NotImplementedError("OSS has no created timestamp")
        bucket_info = self._call_oss("get_bucket_info", bucket=bucket_name)
        timestamp = bucket_info.creation_date
        return datetime.fromtimestamp(timestamp)

    def modified(self, path):
        """Return the modified timestamp of a file as a datetime.datetime"""
        bucket_name, obj_name = self.split_path(path)
        if not obj_name or self.isdir(path):
            raise NotImplementedError("bucket has no modified timestamp")
        simplifiedmeta = self._call_oss("get_object_meta", obj_name, bucket=bucket_name)
        return int(
            datetime.strptime(
                simplifiedmeta.headers["Last-Modified"],
                "%a, %d %b %Y %H:%M:%S %Z",
            ).timestamp()
        )

    def append_object(self, path: str, location: int, value: bytes) -> int:
        """
        Append bytes to the object
        """
        bucket_name, obj_name = self.split_path(path)
        result = self._call_oss(
            "append_object",
            obj_name,
            location,
            value,
            bucket=bucket_name,
        )
        return result.next_position

    def get_object(self, path: str, start: int, end: int) -> bytes:
        """
        Return object bytes in range
        """
        headers = {"x-oss-range-behavior": "standard"}
        bucket_name, obj_name = self.split_path(path)
        try:
            object_stream = self._call_oss(
                "get_object",
                obj_name,
                bucket=bucket_name,
                byte_range=(start, end),
                headers=headers,
            )
        except oss2.exceptions.ServerError as err:
            raise err
        return object_stream.read()

    def sign(self, path, expiration=100, **kwargs):
        raise NotImplementedError("Sign is not implemented for this filesystem")

    def touch(self, path, truncate=True, **kwargs):
        """Create empty file, or update timestamp

        Parameters
        ----------
        path: str
            file location
        truncate: bool
            If True, always set file size to 0; if False, update timestamp and
            leave file unchanged, if backend allows this
        """
        if truncate or not self.exists(path):
            with self.open(path, "wb", **kwargs):
                pass
            self.invalidate_cache(self._parent(path))

    def pipe_file(self, path, value, **kwargs):
        """Set the bytes of given file"""
        bucket_name, obj_name = self.split_path(path)
        self._call_oss("put_object", obj_name, value, bucket=bucket_name, **kwargs)
        bucket = self._get_bucket(bucket_name)
        bucket.put_object(obj_name, value, **kwargs)
        self.invalidate_cache(self._parent(path))

    def invalidate_cache(self, path=None):
        if path is None:
            self.dircache.clear()
        else:
            path = self._strip_protocol(path)
            path = path.lstrip("/")
            self.dircache.pop(path, None)
            while path:
                self.dircache.pop(path, None)
                path = self._parent(path)


class OSSFile(AbstractBufferedFile):
    """A file living in OSSFileSystem"""

    def _upload_chunk(self, final=False):
        """Write one part of a multi-block file upload
        Parameters
        ==========
        final: bool
            This is the last block, so should complete file, if
            self.autocommit is True.
        """
        self.loc = self.fs.append_object(self.path, self.loc, self.buffer.getvalue())
        return True

    def _initiate_upload(self):
        """Create remote file/upload"""
        if "a" in self.mode:
            self.loc = 0
            if self.fs.exists(self.path):
                self.loc = self.fs.info(self.path)["size"]
        elif "w" in self.mode:
            # create empty file to append to
            self.loc = 0
            if self.fs.exists(self.path):
                self.fs.rm_file(self.path)

    def _fetch_range(self, start, end):
        """
        Get the specified set of bytes from remote
        Parameters
        ==========
        start: int
        end: int
        """
        start = max(start, 0)
        end = min(self.size, end)
        if start >= end or start >= self.size:
            return b""
        return self.fs.get_object(self.path, start, end)



"""
Test class level functionality.
"""
# pylint:disable=protected-access
# pylint:disable=missing-function-docstring
# pylint:disable=invalid-name
import pickle
import time
from multiprocessing.pool import ThreadPool

import fsspec.core
import pytest

from ossfs import OSSFileSystem


@pytest.mark.parametrize("default_cache_type", ["none", "bytes", "readahead"])
def test_default_cache_type(init_config, default_cache_type, test_path):
    data = b"a" * (10 * 2**20)
    file = test_path + "/test_default_cache_type/file"
    init_config["default_cache_type"] = default_cache_type
    ossfs = OSSFileSystem(**init_config)
    with ossfs.open(file, "wb") as f:
        f.write(data)

    with ossfs.open(file, "rb") as f:
        assert isinstance(f.cache, fsspec.core.caches[default_cache_type])
        out = f.read(len(data))
        assert len(data) == len(out)
        assert out == data


@pytest.mark.parametrize("cache_type", ["none", "bytes", "readahead"])
def test_cache_type(ossfs, cache_type, test_path):
    data = b"a" * (10 * 2**20)
    file = test_path + "/test_cache_type/file"

    with ossfs.open(file, "wb") as f:
        f.write(data)

    with ossfs.open(file, "rb", cache_type=cache_type) as f:
        assert isinstance(f.cache, fsspec.core.caches[cache_type])
        out = f.read(len(data))
        assert len(data) == len(out)
        assert out == data


def test_current(ossfs, init_config):
    ossfs._cache.clear()  # pylint: disable=protected-access
    ossfs = OSSFileSystem(**init_config)
    assert ossfs.current() is ossfs
    assert OSSFileSystem.current() is ossfs


def test_connect_many(init_config, test_bucket_name):
    def task(num):  # pylint: disable=unused-argument
        ossfs = OSSFileSystem(**init_config)
        ossfs.ls(test_bucket_name)
        time.sleep(5)
        ossfs.ls(test_bucket_name)
        return True

    pool = ThreadPool(processes=20)
    out = pool.map(task, range(40))
    assert all(out)
    pool.close()
    pool.join()


def test_pickle(ossfs, test_path):
    path = test_path + "/test_pickle/"
    for number in range(10):
        ossfs.touch(path + "file" + str(number))

    ossfs1 = pickle.loads(pickle.dumps(ossfs))
    assert ossfs.ls(path) == ossfs1.ls(path)
    ossfs2 = pickle.loads(pickle.dumps(ossfs1))
    assert ossfs.ls(path) == ossfs2.ls(path)


def test_strip_protocol():
    """
    Test protocols
    """
    address = "http://oss-cn-hangzhou.aliyuncs.com/mybucket/myobject"
    assert OSSFileSystem._strip_protocol(address) == "/mybucket/myobject"
    address = "oss://mybucket/myobject"
    assert OSSFileSystem._strip_protocol(address) == "/mybucket/myobject"
